from typing import Dict, List, Union, Tuple

from bolinette import types
from bolinette.blnt.database.engines import DatabaseEngine

MappingPyTyping = List[Union['types.mapping.Column', 'types.mapping.Field',
                             'types.mapping.List', 'types.mapping.Definition']]
MappingListPyTyping = Union[MappingPyTyping, Tuple[str, MappingPyTyping]]


class Model:
    __blnt__: 'ModelMetadata' = None

    def __init__(self, database: 'DatabaseEngine'):
        self.__props__ = ModelProps(self, database)

    @classmethod
    def payloads(cls) -> MappingListPyTyping:
        pass

    @classmethod
    def responses(cls) -> MappingListPyTyping:
        pass

    def __repr__(self):
        return f'<Model {self.__blnt__.name}>'


class ModelMetadata:
    def __init__(self, name: str, database: str):
        self.name = name
        self.database = database


class ModelProps:
    def __init__(self, model, database: 'DatabaseEngine'):
        self.model = model
        self.database = database

    def _get_attribute_of_type(self, attr_type):
        return dict([(name, attribute)
                     for name, attribute in vars(self.model.__class__).items()
                     if isinstance(attribute, attr_type)])

    def get_columns(self) -> Dict[str, 'types.defs.Column']:
        return self._get_attribute_of_type(types.defs.Column)

    def get_relationships(self) -> Dict[str, 'types.defs.Relationship']:
        return self._get_attribute_of_type(types.defs.Relationship)

    def get_properties(self) -> Dict[str, 'ModelProperty']:
        return self._get_attribute_of_type(ModelProperty)


class ModelProperty:
    def __init__(self, name, function):
        self.name = name
        self.function = function

    def __repr__(self):
        return f'<ModelProperty {self.name}>'



class Commands:
    def __init__(self):
        self.commands = {}

    def register(self, name, func):
        self.commands[name] = func


commands = Commands()


def command(name):
    def inner(func):
        commands.register(name, func)
        return func

    return inner



from bolinette import blnt
from bolinette.commands import command


@command('init_db')
async def init_db(context: 'blnt.BolinetteContext', *, run_seeders=True, **_):
    await context.db.drop_all()
    await context.db.create_all()
    if run_seeders:
        await context.db.run_seeders(context)



import subprocess


def run_command(command, callback):
    process = subprocess.Popen(command.split(' '), stdout=subprocess.PIPE)
    while True:
        output = process.stdout.readline()
        if not len(output) and process.poll() is not None:
            break
        if output:
            callback(output.strip().decode())



import json_logging
import logging
import sys


class Logger:
    def __new__(cls):
        json_logging.init_non_web(enable_json=True)
        logger = logging.getLogger("test-logger")
        logger.setLevel(logging.DEBUG)
        logger.addHandler(logging.StreamHandler(sys.stdout))
        return logger



import inspect
from typing import Type, Callable, List, Union

from bolinette import blnt, core, web


def model(model_name: str, database: str = 'default'):
    def decorator(model_cls: Type['core.Model']):
        model_cls.__blnt__ = core.ModelMetadata(model_name, database)
        blnt.cache.models[model_name] = model_cls
        return model_cls
    return decorator


def model_property(function):
    return core.ModelProperty(function.__name__, function)


def mixin(mixin_name: str):
    def decorator(mixin_cls: Type['core.Mixin']):
        blnt.cache.mixins[mixin_name] = mixin_cls
        return mixin_cls
    return decorator


def with_mixin(mixin_name: str):
    def decorator(model_cls):
        mixin_cls = blnt.cache.mixins.get(mixin_name)
        for col_name, col_def in mixin_cls.columns().items():
            setattr(model_cls, col_name, col_def)
        for rel_name, rel_def in mixin_cls.relationships(model_cls).items():
            setattr(model_cls, rel_name, rel_def)
        return model_cls
    return decorator


def init_func(func: Callable[['blnt.BolinetteContext'], None]):
    blnt.cache.init_funcs.append(func)
    return func


def seeder(func):
    blnt.cache.seeders.append(func)
    return func


def service(service_name: str, *, model_name: str = None):
    def decorator(service_cls: Type[Union['core.Service', 'core.SimpleService']]):
        service_cls.__blnt__ = core.ServiceMetadata(service_name, model_name or service_name)
        blnt.cache.services[service_name] = service_cls
        return service_cls
    return decorator


def controller(controller_name: str, path: str = None, *,
               namespace: str = '/api', use_service: bool = True,
               service_name: str = None, middlewares: Union[str, List[str]] = None):
    if path is None:
        path = f'/{controller_name}'
    if service_name is None:
        service_name = controller_name
    if middlewares is None:
        middlewares = []
    if isinstance(middlewares, str):
        middlewares = [middlewares]

    def decorator(controller_cls: Type['web.Controller']):
        controller_cls.__blnt__ = web.ControllerMetadata(
            controller_name, path, use_service, service_name, namespace, middlewares)
        blnt.cache.controllers[controller_name] = controller_cls
        return controller_cls
    return decorator


def route(path: str, *, method: web.HttpMethod, expects: 'web.Expects' = None, returns: 'web.Returns' = None,
          middlewares: Union[str, List[str]] = None):
    if middlewares is None:
        middlewares = []
    if isinstance(middlewares, str):
        middlewares = [middlewares]

    def decorator(route_function: Callable):
        if not isinstance(route_function, web.ControllerRoute) and not inspect.iscoroutinefunction(route_function):
            raise ValueError(f'Route "{route_function.__name__}" must be an async function')
        if expects is not None and not isinstance(expects, web.Expects):
            raise ValueError(f'Route "{route_function.__name__}": expects argument must be of type web.Expects')
        if returns is not None and not isinstance(returns, web.Returns):
            raise ValueError(f'Route "{route_function.__name__}": expects argument must be of type web.Returns')
        inner_route = None
        if isinstance(route_function, web.ControllerRoute):
            inner_route = route_function
            route_function = route_function.func
        return web.ControllerRoute(route_function, path, method, expects, returns, inner_route, middlewares)
    return decorator


def get(path: str, *, expects: 'web.Expects' = None, returns: 'web.Returns' = None,
        middlewares: Union[str, List[str]] = None):
    return route(path, method=web.HttpMethod.GET, expects=expects, returns=returns, middlewares=middlewares)


def post(path: str, *, expects: 'web.Expects' = None, returns: 'web.Returns' = None,
         middlewares: Union[str, List[str]] = None):
    return route(path, method=web.HttpMethod.POST, expects=expects, returns=returns, middlewares=middlewares)


def put(path: str, *, expects: 'web.Expects' = None, returns: 'web.Returns' = None,
        middlewares: Union[str, List[str]] = None):
    return route(path, method=web.HttpMethod.PUT, expects=expects, returns=returns, middlewares=middlewares)


def patch(path: str, *, expects: 'web.Expects' = None, returns: 'web.Returns' = None,
          middlewares: Union[str, List[str]] = None):
    return route(path, method=web.HttpMethod.PATCH, expects=expects, returns=returns, middlewares=middlewares)


def delete(path: str, *, expects: 'web.Expects' = None, returns: 'web.Returns' = None,
           middlewares: Union[str, List[str]] = None):
    return route(path, method=web.HttpMethod.DELETE, expects=expects, returns=returns, middlewares=middlewares)


def middleware(name: str, *, priority: int = 10, auto_load: bool = False, loadable: bool = True):
    def decorator(middleware_cls: Type['web.Middleware']):
        middleware_cls.__blnt__ = web.MiddlewareMetadata(name, priority, auto_load, loadable)
        blnt.cache.middlewares[name] = middleware_cls
        return middleware_cls
    return decorator


def topic(topic_name: str):
    def decorator(topic_cls: Type['web.Topic']):
        topic_cls.__blnt__ = web.TopicMetadata(topic_name)
        blnt.cache.topics[topic_name] = topic_cls
        return topic_cls
    return decorator


def channel(rule: str):
    def decorator(channel_function: Callable):
        return web.TopicChannel(channel_function, rule)
    return decorator



import asyncio
import sys
import traceback
from asyncio import AbstractEventLoop
from datetime import datetime
from typing import List

from bolinette import console, blnt
from bolinette import bolinette
from bolinette.testing import TestClient, Bolitest


class TestResult:
    def __init__(self, name: str, ok: bool, time: float, error: Exception = None, tb: str = None):
        self.name = name
        self.ok = ok
        self.time = time
        self.error = error
        self.traceback = tb


class TestRunner:
    def __init__(self, context: 'blnt.BolinetteContext', run_only: str):
        self.context = context
        self.run_only = run_only

    def run_tests(self, blnt_app: 'bolinette.Bolinette'):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._run_tests(blnt_app, loop, self._gather_tests()))

    def _gather_tests(self):
        tests = []
        for test in blnt.cache.test_funcs:
            test.set_name(self.context.root_path('tests'))
            if self._match_test(test):
                tests.append(test)
        return tests

    def _match_test(self, test: Bolitest):
        if self.run_only is None:
            return True
        if '::' in self.run_only:
            return self.run_only == test.name
        return test.name.startswith(self.run_only)

    async def _run_tests(self, blnt_app: 'bolinette.Bolinette', loop: AbstractEventLoop, tests: List[Bolitest]):
        tests_start_time = datetime.now()
        test_cnt = len(tests)
        console.print('** Bolinette API Tests **')
        console.print(f'Running {test_cnt} tests ({self.run_only or "all tests"}), starting at {tests_start_time}')
        console.print('====================\n')
        results = []
        test_index = 0
        for test in tests:
            client = TestClient(blnt_app, loop)
            async with client:
                console.print(f'Running: {test.name} [{test_index + 1}/{test_cnt}]', end=' ')
                result = await self._run_test(client, test)
                if result.ok:
                    console.print(f'OK in {result.time / 1000}ms')
                else:
                    console.print(f'FAILED ({result.error.__class__.__name__}) in {result.time / 1000}ms')
                results.append(result)
            test_index += 1
        tests_end_time = datetime.now()

        err_cnt = 0
        for result in results:
            if result.error:
                err_cnt += 1
                console.error(f'\n========== ERROR: {result.name} ==========')
                console.error(result.traceback, end='')

        await asyncio.sleep(.1)
        console.print('\n====================')
        console.print(f'Ran {test_cnt} tests with {err_cnt} errors '
                      f'in {(tests_end_time - tests_start_time).seconds}s')
        if err_cnt > 0:
            sys.exit(1)

    async def _run_test(self, client: TestClient, test: Bolitest) -> TestResult:
        time_start = datetime.now()
        try:
            await test.func(client)
            time_end = datetime.now()
            return TestResult(test.name, True, (time_end - time_start).microseconds)
        except Exception as err:
            time_end = datetime.now()
            return TestResult(test.name, False, (time_end - time_start).microseconds,
                              err, traceback.format_exc())



import asyncio
import inspect

from aiohttp import web as aio_web
from bolinette.utils import paths

from bolinette import blnt, console
from bolinette.commands import commands
from bolinette.exceptions import InitError
from bolinette.utils.functions import invoke


class Bolinette:
    def __init__(self, *, profile=None, overrides=None):
        try:
            self.app = aio_web.Application()
            self.context = blnt.BolinetteContext(paths.dirname(__file__), self.app,
                                                 profile=profile, overrides=overrides)
            self.app['blnt'] = self.context
            self.run_init_functions(self.app)
        except InitError as init_error:
            console.error(f'Error raised during Bolinette init phase\n{str(init_error)}')
            exit(1)

    @staticmethod
    def run_init_functions(app):
        for func in blnt.cache.init_funcs:
            func(app['blnt'])

    def run(self):
        self.context.logger.info(f"Starting Bolinette with '{self.context.env['profile']}' environment profile")
        aio_web.run_app(self.app, port=self.context.env['port'], access_log=self.context.logger)
        self.context.logger.info(f"Bolinette stopped gracefully")

    def run_command(self, *args, **kwargs):
        kwargs['blnt'] = self
        name = args[0]
        kwargs['args'] = args[1:]
        if name in commands.commands:
            func = commands.commands[name]
            if inspect.isfunction(func):
                if inspect.iscoroutinefunction(func):
                    loop = asyncio.get_event_loop()
                    loop.run_until_complete(func(self.context, **kwargs))
                else:
                    invoke(func, **kwargs)



import sqlalchemy

from bolinette import blnt, core
from bolinette.blnt.database.engines import RelationalDatabase
from bolinette.blnt.database.queries import BaseQueryBuilder, BaseQuery


class RelationalQueryBuilder(BaseQueryBuilder):
    def __init__(self, model: 'core.Model', context: 'blnt.BolinetteContext'):
        super().__init__(model, context)
        self._database: RelationalDatabase = context.db[model.__blnt__.database]
        self._table = self.context.table(model.__blnt__.name)

    def query(self) -> 'BaseQuery':
        return RelationalQuery(self._database, self._table)

    async def insert_entity(self, values):
        entity = self._table(**values)
        self._database.session.add(entity)
        return entity

    async def delete_entity(self, entity):
        self._database.session.delete(entity)
        return entity


class RelationalQuery(BaseQuery):
    def __init__(self, database: RelationalDatabase, table):
        super().__init__()
        self._database = database
        self._table = table

    def _clone(self):
        query = RelationalQuery(self._database, self._table)
        self._base_clone(query)
        return query

    async def all(self):
        return self._build_query().all()

    async def first(self):
        return self._build_query().first()

    async def get_by_id(self, identifier):
        return self._query().get(identifier)

    async def count(self):
        return self._build_query().count()

    def _query(self):
        return self._database.session.query(self._table)

    def _build_query(self):
        query = self._query()
        if len(self._filters) > 0:
            query = query.filter_by(**self._filters)
        if len(self._order_by):
            orders = []
            for func, desc in self._order_by:
                col = func(self._table)
                if desc:
                    col = sqlalchemy.desc(col)
                orders.append(col)
        return query



from bolinette.testing.mock import Mock, Mocked
from bolinette.testing.client import TestClient
from bolinette.testing.test import Bolitest
from bolinette.testing.decorator import bolitest
from bolinette.testing.runner import TestRunner



from bolinette import types, core, mapping
from bolinette.decorators import model


@model('role')
class Role(core.Model):
    id = types.defs.Column(types.db.Integer, primary_key=True)
    name = types.defs.Column(types.db.String, unique=True, nullable=False)

    @classmethod
    def payloads(cls):
        yield [
            mapping.Column(cls.name, required=True)
        ]

    @classmethod
    def responses(cls):
        yield [
            mapping.Column(cls.name)
        ]
        yield 'complete', [
            mapping.Column(cls.name),
            mapping.List(mapping.Definition('user'), key='users')
        ]



import inspect
from functools import wraps
from typing import Callable, Awaitable, Any

from bolinette import blnt
from bolinette.testing import TestClient, Bolitest
from bolinette.utils.functions import async_invoke


def bolitest(*, before: Callable[[Any], Awaitable[None]] = None,
             after: Callable[[Any], Awaitable[None]] = None):
    def wrapper(func):
        @wraps(func)
        async def inner(client: TestClient):
            try:
                await client.context.db.drop_all()
                await client.context.db.create_all()
                if before is not None:
                    await async_invoke(before, context=client.context, mock=client.mock)
                await client.context.db.close_transaction()
                await func(client=client)
                if after is not None:
                    await async_invoke(after, context=client.context, mock=client.mock)
                await client.context.db.drop_all()
            except Exception:
                await client.context.db.rollback_transaction()
        blnt.cache.test_funcs.append(Bolitest(inner, inspect.getfile(func)))
        return inner
    return wrapper



import requests
from bolinette.utils import logger

from bolinette import blnt


class Mailgun:
    def __init__(self, context: 'blnt.BolinetteContext'):
        self.url = context.env['mailgun_url']
        self.ready = True
        self.key = context.env['mailgun_api']
        self.from_adr = context.env['mailgun_from']
        self._validate_attrs()

    def _validate_attrs(self):
        if not self.url:
            self.ready = False
            logger.error('MAILGUN_URL env key not set')
        if not self.key:
            self.ready = False
            logger.error('MAILGUN_API env key not set')
        if not self.from_adr:
            self.ready = False
            logger.error('MAILGUN_FROM env key not set')

    def send(self, to, subject, content):
        if self.ready:
            return requests.post(self.url, auth=("api", self.key), data={
                'from': self.from_adr,
                'to': list(to),
                'subject': subject,
                'text': content
            })



from bolinette.web.method import HttpMethod
from bolinette.web.controller import (
    Controller, ControllerMetadata, ControllerRoute, Expects, Returns, MiddlewareTrack
)
from bolinette.web.middleware import Middleware, MiddlewareMetadata, InternalMiddleware
from bolinette.web.topic import Topic, TopicMetadata, TopicChannel
from bolinette.web.resources import BolinetteResources, BolinetteResource
from bolinette.web.sockets import BolinetteSockets
from bolinette.web.response import APIResponse, Response, Cookie



from aiohttp import web as aio_web

from bolinette import web
from bolinette.decorators import topic, channel


@topic('user')
class UserTopic(web.Topic):
    @channel(r'echo')
    async def message(self, socket: aio_web.WebSocketResponse, payload, current_user, **_):
        await socket.send_str(f'{payload["message"]} from {current_user.username if current_user else "anonymous"}')
        await self.context.sockets.send_message('user', 'root', 'hey root, what\'s up?')



from bson import ObjectId
from pymongo.collection import Collection

from bolinette import blnt, core
from bolinette.blnt.database.engines import CollectionDatabase
from bolinette.blnt.database.queries import BaseQueryBuilder, BaseQuery


class CollectionQueryBuilder(BaseQueryBuilder):
    def __init__(self, model: 'core.Model', context: 'blnt.BolinetteContext'):
        super().__init__(model, context)
        self._name = model.__blnt__.name
        self._database: 'CollectionDatabase' = context.db[model.__blnt__.database]
        self._collection = self._database.db[self._name]

    def query(self) -> 'BaseQuery':
        return CollectionQuery(self._collection)

    async def insert_entity(self, values):
        result = self._collection.insert_one(values)
        return self._collection.find_one({'_id': ObjectId(result.inserted_id)})

    async def delete_entity(self, entity):
        self._collection.delete_one({'_id': entity['_id']})
        return entity


class CollectionQuery(BaseQuery):
    def __init__(self, collection: Collection):
        super().__init__()
        self._collection = collection

    def _clone(self):
        query = CollectionQuery(self._collection)
        self._base_clone(query)
        return query

    async def all(self):
        return self._collection.find(self._build_filters())

    async def first(self):
        return self._collection.find_one(self._build_filters())

    async def get_by_id(self, identifier):
        return self._collection.find_one({'_id': ObjectId(identifier)})

    async def count(self):
        pass

    def _build_filters(self):
        params = {}
        if len(self._filters) > 0:
            for key in self._filters:
                if key == '_id':
                    params[key] = ObjectId(self._filters[key])
                else:
                    params[key] = self._filters[key]
        return params



from dateutil import parser as date_parser

from bolinette import blnt, types, exceptions, mapping
from bolinette.exceptions import APIErrors, APIError, EntityNotFoundError


class Validator:
    def __init__(self, context: 'blnt.BolinetteContext'):
        self.context = context

    def validate_payload(self, model: str, key: str, values, patch=False):
        api_errors = APIErrors()
        valid = {}
        definition = self.context.mapper.payload(model, key)
        for field in definition.fields:
            field_key = None
            field_name = None
            if isinstance(field, mapping.Field):
                field_key = field.key
                field_name = field.name
            elif isinstance(field, mapping.Reference):
                field_key = field.foreign_key
                field_name = field.foreign_key
            if patch and field_name not in values:
                continue
            if field_key is not None and field_name is not None:
                try:
                    valid[field_key] = self._validate_field(field, field_name, values)
                except APIError as ex:
                    api_errors.append(ex)
        if api_errors:
            raise api_errors
        return valid

    def _validate_field(self, field, name, values):
        if field.required and name not in values:
            raise exceptions.ParamMissingError(name)
        value = values.get(name, field.default)
        if not value and not field.nullable:
            raise exceptions.ParamNonNullableError(name)
        if value is not None and isinstance(field, mapping.Field):
            if field.type == types.db.Date:
                value = date_parser.parse(value)
        return value

    async def link_foreign_entities(self, model: str, key: str, params):
        api_errors = APIErrors()
        definition = self.context.mapper.payload(model, key)
        for field in definition.fields:
            if isinstance(field, mapping.Reference):
                value = params.get(field.foreign_key, None)
                repo = self.context.repo(field.reference_model)
                if value is not None and repo is not None:
                    entity = await repo.get_first_by(field.reference_key, value)
                    if entity is None:
                        api_errors.append(EntityNotFoundError(field.reference_model, field.reference_key, value))
        if api_errors:
            raise api_errors



from typing import Callable, Awaitable

from bolinette.testing import TestClient


class Bolitest:
    def __init__(self, func: Callable[[TestClient], Awaitable[None]], file: str):
        self.func = func
        self.name = None
        self._func_name = func.__name__
        self._file = file

    def set_name(self, root_path):
        self.name = self._file.replace(root_path + '/', '').replace('.py', '') + '::' + self._func_name



from bolinette.defaults.services.timezone import TimezoneService
from bolinette.defaults.services.file import FileService
from bolinette.defaults.services.role import RoleService
from bolinette.defaults.services.user import UserService



import mimetypes

from aiohttp import web as aio_web
from bolinette.utils import paths

from bolinette import web
from bolinette.decorators import get, controller


@controller('static', namespace='', use_service=False)
class StaticController(web.Controller):
    async def _file_sender(self, path):
        with open(path, 'rb') as file:
            chunk = file.read(2 ** 16)
            while chunk:
                yield chunk
                chunk = file.read(2 ** 16)

    @get('/{route:.*}')
    async def get_static_file(self, match):
        route = match['route'].split('/')
        if route == ['']:
            route = ['index.html']
        path = self.context.static_path(*route)
        if paths.exists(path):
            content_type = mimetypes.guess_type(path)[0]
            return aio_web.Response(body=self._file_sender(path),
                                    status=200, content_type=content_type or 'text/plain')
        return aio_web.Response(body='global.404_error', status=404, content_type='text/plain')



from bolinette.core import functions
from bolinette.core.model import Model, ModelMetadata, ModelProperty
from bolinette.core.mixin import Mixin
from bolinette.core.repository import Repository
from bolinette.core.service import Service, ServiceMetadata, SimpleService, HistorizedService



from datetime import datetime
from typing import List

from bolinette import blnt, core
from bolinette.blnt.objects import PaginationParams, OrderByParams
from bolinette.exceptions import EntityNotFoundError


class Service:
    __blnt__: 'ServiceMetadata' = None

    def __init__(self, context: 'blnt.BolinetteContext'):
        self.context = context

    def __repr__(self):
        return f'<Service {self.__blnt__.name}>'

    @property
    def repo(self) -> core.Repository:
        return self.context.repo(self.__blnt__.model_name)

    async def get(self, identifier, *, safe=False):
        entity = await self.repo.get(identifier)
        if entity is None and not safe:
            raise EntityNotFoundError(model=self.__blnt__.name, key='id', value=identifier)
        return entity

    async def get_by(self, key, value):
        return await self.repo.get_by(key, value)

    async def get_first_by(self, key, value, *, safe=False):
        entity = await self.repo.get_first_by(key, value)
        if entity is None and not safe:
            raise EntityNotFoundError(model=self.__blnt__.name, key=key, value=value)
        return entity

    async def get_all(self, pagination: PaginationParams = None, order_by: List[OrderByParams] = None):
        return await self.repo.get_all(pagination, order_by)

    async def create(self, values):
        return await self.repo.create(values)

    async def update(self, entity, values):
        return await self.repo.update(entity, values)

    async def patch(self, entity, values):
        return await self.repo.patch(entity, values)

    async def delete(self, entity):
        return await self.repo.delete(entity)


class SimpleService:
    __blnt__: 'ServiceMetadata' = None

    def __init__(self, context: 'blnt.BolinetteContext'):
        self.context = context

    def __repr__(self):
        return f'<Service {self.__blnt__.name}>'


class ServiceMetadata:
    def __init__(self, name: str, model_name: str):
        self.name = name
        self.model_name = model_name


class HistorizedService(Service):
    def __init__(self, context: 'blnt.BolinetteContext'):
        super().__init__(context)

    async def create(self, values, *, current_user=None):
        if current_user:
            now = datetime.utcnow()
            values['created_on'] = now
            values['created_by_id'] = current_user.id
            values['updated_on'] = now
            values['updated_by_id'] = current_user.id
        return await super().create(values)

    async def update(self, entity, values, *, current_user=None):
        if current_user:
            now = datetime.utcnow()
            values['created_on'] = entity.created_on
            values['created_by_id'] = entity.created_by_id
            values['updated_on'] = now
            values['updated_by_id'] = current_user.id
        return await super().update(entity, values)

    async def patch(self, entity, values, *, current_user=None):
        if current_user:
            now = datetime.utcnow()
            values['updated_on'] = now
            values['updated_by_id'] = current_user.id
        return await super().patch(entity, values)



import inspect
from typing import Callable, Awaitable, Any

from bolinette.exceptions import InternalError


def _parse_params(function, *args, **kwargs):
    cur_arg = 0
    func_params = {}
    for key, param in inspect.signature(function).parameters.items():
        if param.kind == param.VAR_KEYWORD:
            for name, value in kwargs.items():
                if name not in func_params:
                    func_params[name] = value
        elif param.kind in [param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD] and cur_arg < len(args):
            func_params[key] = args[cur_arg]
            cur_arg += 1
        elif key in kwargs:
            func_params[key] = kwargs[key]
        else:
            func_params[key] = None
    return func_params


async def async_invoke(function: Callable[[Any], Awaitable[Any]], *args, **kwargs):
    if inspect.iscoroutinefunction(function):
        return await function(**_parse_params(function, *args, **kwargs))
    raise InternalError(f'internal.not_async_function:{function.__name__}')


def invoke(function: Callable[[Any], Any], *args, **kwargs):
    if inspect.isfunction(function):
        return function(**_parse_params(function, *args, **kwargs))
    raise InternalError(f'internal.not_function:{function.__name__}')


def getattr_(entity, key, default):
    if isinstance(entity, dict):
        return entity.get(key, default)
    return getattr(entity, key, default)


def hasattr_(entity, key):
    if isinstance(entity, dict):
        return key in entity
    return hasattr(entity, key)


def setattr_(entity, key, value):
    if isinstance(entity, dict):
        entity[key] = value
    else:
        setattr(entity, key, value)



from typing import Dict

from bolinette import blnt
from bolinette.blnt.database.engines import RelationalDatabase, CollectionDatabase, DatabaseEngine
from bolinette.exceptions import InternalError, InitError


class DatabaseManager:
    _DBMS = {
        'sqlite://': RelationalDatabase,
        'mongodb://': CollectionDatabase
    }

    def __init__(self, context: 'blnt.BolinetteContext'):
        self.context = context
        self.engines: Dict[str, DatabaseEngine] = {}
        self._init_databases()

    def _init_databases(self):
        def _init_database(_name: str, _uri: str):
            for dbms in self._DBMS:
                if _uri.startswith(dbms):
                    self.engines[_name] = self._DBMS[dbms](_uri)
                    break
            else:
                raise InitError(f'Unsupported database system for URI "{_uri}"')
        try:
            conf = self.context.env['database']
            if isinstance(conf, str):
                _init_database('default', conf)
            elif isinstance(conf, dict):
                for name, uri in conf.items():
                    if not isinstance(uri, str):
                        raise ValueError()
                    _init_database(name, uri)
            else:
                raise ValueError()
        except ValueError:
            raise InitError('Bad database configuration in env files')

    @property
    def engine(self):
        if 'default' in self.engines:
            return self.engines['default']
        raise InternalError('internal.db.no_default_engine')

    def __getitem__(self, key):
        if key in self.engines:
            return self.engines[key]
        raise InternalError(f'internal.db.no_engine:{key}')

    def __contains__(self, key):
        return key in self.engines

    async def open_transaction(self):
        for _, engine in self.engines.items():
            await engine.open_transaction()

    async def close_transaction(self):
        for _, engine in self.engines.items():
            await engine.close_transaction()

    async def rollback_transaction(self):
        for _, engine in self.engines.items():
            await engine.rollback_transaction()

    async def create_all(self):
        for _, engine in self.engines.items():
            await engine.create_all()

    async def drop_all(self):
        for _, engine in self.engines.items():
            await engine.drop_all()

    async def run_seeders(self, context: 'blnt.BolinetteContext'):
        for func in blnt.cache.seeders:
            await func(context)



import bolinette
from bolinette import testing
from bolinette.commands import command


@command('run_tests')
def run_tests(blnt: 'bolinette.Bolinette', args):
    runner = testing.TestRunner(blnt.context, args[0] if len(args) > 0 else None)
    runner.run_tests(blnt)



import os
import random
import string
from typing import Dict, Any

import yaml

from bolinette import console, blnt
from bolinette.exceptions import InitError
from bolinette.utils import paths


class Settings:
    def __init__(self):
        self._settings: Dict[str, Any] = {}

    def __getitem__(self, key: str):
        return self._settings.get(key.lower(), None)

    def __contains__(self, key: str):
        return key.lower() in self._settings

    def __setitem__(self, key: str, value):
        self._settings[key.lower()] = value

    def __repr__(self):
        return repr(self._settings)

    def get(self, key: str, default=None):
        item = self[key]
        return item if item is not None else default

    def _cwd_path(self, *path):
        return paths.join(paths.cwd(), *path)

    def _reset(self, settings: Dict[str, Any]):
        self._settings = {}
        for key, value in settings.items():
            self._settings[key.lower()] = value

    def _load_from_file(self, file_name):
        try:
            with open(self._cwd_path('env', file_name), 'r') as f:
                d = yaml.safe_load(f)
                if not isinstance(d, dict):
                    raise yaml.YAMLError()
                return d
        except FileNotFoundError:
            return {}
        except yaml.YAMLError:
            raise InitError(f'File "{file_name}" does not contain valid YAML syntax')


class InitSettings(Settings):
    def __init__(self):
        super().__init__()
        self._DEFAULT_PROFILE = 'development'
        self._init_checks()
        self.profile = self._read_profile()
        self._reset(self._load_from_file('init.yaml'))

    def _read_profile(self):
        if 'BLNT_PROFILE' in os.environ:
            return os.environ['BLNT_PROFILE']
        try:
            with open(self._cwd_path('env', '.profile')) as f:
                profile = f.read().split('\n')[0].strip()
                if not profile:
                    console.error(f'Warning: empty "env/.profile", defaulting to "{self._DEFAULT_PROFILE}"')
                    return self._DEFAULT_PROFILE
                return profile
        except FileNotFoundError:
            console.error(f'Warning: no "env/.profile" file, defaulting to "{self._DEFAULT_PROFILE}"')
            return self._DEFAULT_PROFILE

    def _init_checks(self):
        dirs = [('env',), ('instance',)]
        for d in dirs:
            if not paths.exists(self._cwd_path(*d)):
                console.error(f'Warning: no "{"/".join(d)}" dir, creating one')
                paths.mkdir(self._cwd_path(*d))


init = InitSettings()


class Environment(Settings):
    def __init__(self, context: 'blnt.BolinetteContext', *, profile=None, overrides=None):
        super().__init__()
        self.context = context
        profile = profile or init.profile or 'development'
        self._reset(self._merge_env_stack([
            self._default_env,
            self._load_from_file(f'env.{profile}.yaml'),
            self._load_from_file(f'env.local.{profile}.yaml'),
            self._load_from_os(),
            overrides or {},
            {'profile': profile}
        ]))
        self._check_secret_key()

    @property
    def _default_env(self):
        return {
            'app_name': 'default_name',
            'app_desc': 'default_description',
            'app_version': '0.0.1',
            'database': None,
            'debug': False,
            'host': '127.0.0.1',
            'port': '5000',
            'webapp_folder': self.context.root_path('webapp', 'dist')
        }

    def _load_from_os(self):
        keys = {}
        for key in os.environ:
            if key.startswith('BLNT_'):
                keys[key[5:].lower()] = os.environ[key]
        return keys

    def _merge_env_stack(self, stack):
        settings = {}
        for source in stack:
            for key, value in source.items():
                settings[key.lower()] = value
        return settings

    def _check_secret_key(self):
        if self['secret_key'] is None:
            console.error('Warning: no secret_key set, using random one')
            self['secret_key'] = ''.join(random.choices(string.ascii_letters + string.digits, k=64))



import jinja2
import yaml
from jinja2 import TemplateNotFound

from bolinette.exceptions import InternalError
from bolinette.utils import paths


def write(path, content, mode='w+'):
    with open(path, mode=mode) as file:
        file.write(content)


def append(path, content):
    write(path, content, mode='a+')


def read_file(path):
    try:
        with open(path) as f:
            return f.read()
    except FileNotFoundError:
        return None


def read_version(path):
    try:
        with open(paths.join(path, '.version')) as f:
            for line in f:
                return line.strip().replace('\n', '')
    except FileNotFoundError:
        return None


def read_requirements(path):
    try:
        with open(paths.join(path, 'requirements.txt')) as f:
            return list(filter(lambda r: len(r), f.read().split('\n')))
    except FileNotFoundError:
        return []


def read_manifest(path):
    try:
        with open(paths.join(path, 'manifest.blnt.yml')) as f:
            return yaml.safe_load(f)
    except FileNotFoundError:
        return None


def render_template(path, params):
    template_path, template_name = paths.split(path)
    if not len(template_path):
        template_path = paths.join(paths.dirname(__file__), 'files')
    jinja_env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(searchpath=template_path),
        keep_trailing_newline=True,
        trim_blocks=True,
        lstrip_blocks=True
    )
    try:
        template = jinja_env.get_template(template_name)
    except TemplateNotFound:
        raise InternalError(f'internal.template.not_found:{path}')
    return template.render(**params)



from typing import Callable, List, Dict, Any

from aiohttp.web_request import Request

from bolinette import blnt, core, web
from bolinette.exceptions import InitError
from bolinette.utils import functions


class Controller:
    __blnt__: 'ControllerMetadata' = None

    def __init__(self, context: 'blnt.BolinetteContext'):
        self.__props__ = ControllerProps(self)
        self.context = context
        self.response = context.response
        if self.__blnt__.use_service:
            self.defaults = ControllerDefaults(self)

    @property
    def service(self) -> 'core.Service':
        return self.context.service(self.__blnt__.service_name)

    def default_routes(self):
        return []

    def __repr__(self):
        return f'<Controller {self.__blnt__.name} {self.__blnt__.path}>'


class ControllerMetadata:
    def __init__(self, name: str, path: str, use_service: bool, service_name: str,
                 namespace: str, middlewares: List[str]):
        self.name = name
        self.path = path
        self.use_service = use_service
        self.service_name = service_name
        self.namespace = namespace
        self.middlewares = middlewares


class ControllerProps:
    def __init__(self, controller):
        self.controller = controller

    def _get_attribute_of_type(self, attr_type):
        return dict([(name, attribute)
                     for name, attribute in vars(self.controller.__class__).items()
                     if isinstance(attribute, attr_type)])

    def get_routes(self) -> Dict[str, 'ControllerRoute']:
        return self._get_attribute_of_type(ControllerRoute)


class ControllerRoute:
    def __init__(self, func: Callable, path: str, method: web.HttpMethod,
                 expects: 'Expects' = None, returns: 'Returns' = None,
                 inner_route: 'ControllerRoute' = None, middlewares: List[str] = None):
        self.controller = None
        self.func = func
        self.path = path
        self.method = method
        self.expects = expects
        self.returns = returns
        self.inner_route = inner_route
        self._mdw_defs = middlewares or []
        self.middlewares: List['web.Middleware'] = []

    def init_middlewares(self, context: 'blnt.BolinetteContext', from_controller: List[str], system: List[str]):
        for mdw in system:
            self._parse_middleware_options(mdw, context, True)
        for mdw in from_controller:
            self._parse_middleware_options(mdw, context)
        for mdw in self._mdw_defs:
            self._parse_middleware_options(mdw, context)
        self.middlewares = sorted(sorted(self.middlewares, key=lambda m: m.__blnt__.priority),
                                  key=lambda m: m.system_priority)

    def _parse_middleware_options(self, mdw: str, context, system: bool = False):
        name, *args = mdw.split('|')
        if name.startswith('!'):
            self.middlewares = list(filter(lambda m: m.__blnt__.name != name[1:], self.middlewares))
            return
        find = list(filter(lambda m: m.__blnt__.name == name, self.middlewares))
        if len(find) > 0:
            middleware = find[0]
            middleware.options = {}
        else:
            middleware = blnt.cache.middlewares[name](context)

        if not middleware.__blnt__.loadable and not system:
            raise InitError(f'[{self.controller.__class__.__name__}] Middleware '
                            f'"{mdw}" cannot be loaded in a controller')

        for arg in args:
            arg_n, *arg_v = arg.split('=', maxsplit=1)
            arg_n, *filters = arg_n.split(':')
            if len(filters) > 1:
                raise InitError(f'[{self.controller.__class__.__name__}] Middleware '
                                f'"{name}|{arg_n}" has too many filters')
            if len(arg_v) == 0:
                value = True
            else:
                value = arg_v[0].split(',')
                if len(value) == 1:
                    value = value[0]
                if len(filters) == 1:
                    value = self._apply_filters(value, filters[0], name, arg_n)
            middleware.options[arg_n] = value
        self.middlewares.append(middleware)

    def _apply_filters(self, value, _filter, mdw, arg):
        if _filter == 'int':
            func = int
        elif _filter == 'float':
            func = float
        else:
            raise InitError(f'[{self.controller.__class__.__name__}] middleware '
                            f'"{mdw}|{arg}": unknown filter "{_filter}"')
        try:
            if isinstance(value, list):
                return [func(v) for v in value]
            return func(value)
        except ValueError:
            raise InitError(f'[{self.controller.__class__.__name__}] middleware '
                            f'"{mdw}|{arg}": unable to parse "{value}" as {_filter}')

    async def call_middleware_chain(self, request: Request, params: Dict[str, Any], track: 'MiddlewareTrack'):
        handles = ([MiddlewareHandle(name=m.__blnt__.name, func=m.handle) for m in self.middlewares]
                   + [MiddlewareHandle('__ctrl_call', self._final_middleware_call)])
        return await self._call_middleware(handles, 0, request, params, track)

    async def _call_middleware(self, handles: List['MiddlewareHandle'], index: int, request: Request,
                               params: Dict[str, Any], track: 'MiddlewareTrack'):
        async def _next(_request, _params):
            return await self._call_middleware(handles, index + 1, _request, _params, track)
        track.append(handles[index].name)
        track.done = handles[index].name == '__ctrl_call'
        return await handles[index].func(request, params, _next)

    async def _final_middleware_call(self, _1, params: Dict[str, Any], _2):
        return await functions.async_invoke(self.func, self.controller, **params)


class MiddlewareHandle:
    def __init__(self, name: str, func: Callable):
        self.name = name
        self.func = func


class MiddlewareTrack:
    def __init__(self):
        self.done = False
        self.steps: List[str] = []

    def append(self, step: str):
        self.steps.append(step)


class Expects:
    def __init__(self, model: str, key: str = 'default', *, patch: bool = False):
        self.model = model
        self.key = key
        self.patch = patch


class Returns:
    def __init__(self, model: str, key: str = 'default', *, as_list: bool = False, skip_none: bool = False):
        self.model = model
        self.key = key
        self.as_list = as_list
        self.skip_none = skip_none


class ControllerDefaults:
    def __init__(self, controller: Controller):
        self.service: core.Service = controller.service

    def get_all(self, returns='default', *, middlewares=None):
        async def route(controller, **kwargs):
            resp = await functions.async_invoke(controller.service.get_all, **kwargs)
            return controller.response.ok('OK', resp)

        return ControllerRoute(route, '', web.HttpMethod.GET,
                               returns=Returns(self.service.__blnt__.model_name, returns, as_list=True),
                               middlewares=middlewares)

    def get_one(self, returns='default', *, key='id', middlewares=None):
        async def route(controller, *, match, **kwargs):
            resp = await functions.async_invoke(controller.service.get_first_by, key, match.get('value'), **kwargs)
            return controller.response.ok('OK', resp)

        return ControllerRoute(route, '/{value}', web.HttpMethod.GET,
                               returns=Returns(self.service.__blnt__.model_name, returns),
                               middlewares=middlewares)

    def create(self, returns='default', expects='default', *, middlewares=None):
        async def route(controller, payload, **kwargs):
            resp = await functions.async_invoke(controller.service.create, payload, **kwargs)
            return controller.response.created(f'{controller.service.__blnt__.model_name}.created', resp)

        return ControllerRoute(route, '', web.HttpMethod.POST,
                               expects=Expects(self.service.__blnt__.model_name, expects),
                               returns=Returns(self.service.__blnt__.model_name, returns),
                               middlewares=middlewares)

    def update(self, returns='default', expects='default', *, key='id', middlewares=None):
        async def route(controller, payload, match, **kwargs):
            entity = await controller.service.get_first_by(key, match.get('value'))
            resp = await functions.async_invoke(controller.service.update, entity, payload, **kwargs)
            return controller.response.ok(f'{controller.service.__blnt__.model_name}.updated', resp)

        return ControllerRoute(route, '/{value}', web.HttpMethod.PUT,
                               expects=Expects(self.service.__blnt__.model_name, expects),
                               returns=Returns(self.service.__blnt__.model_name, returns),
                               middlewares=middlewares)

    def patch(self, returns='default', expects='default', *, key='id', middlewares=None):
        async def route(controller, payload, match, **kwargs):
            entity = await controller.service.get_first_by(key, match.get('value'))
            resp = await functions.async_invoke(controller.service.patch, entity, payload, **kwargs)
            return controller.response.ok(f'{controller.service.__blnt__.model_name}.updated', resp)

        return ControllerRoute(route, '/{value}', web.HttpMethod.PATCH,
                               expects=Expects(self.service.__blnt__.model_name, expects, patch=True),
                               returns=Returns(self.service.__blnt__.model_name, returns),
                               middlewares=middlewares)

    def delete(self, returns='default', *, key='id', middlewares=None):
        async def route(controller, match, **kwargs):
            entity = await controller.service.get_first_by(key, match.get('value'))
            resp = await functions.async_invoke(controller.service.delete, entity, **kwargs)
            return controller.response.ok(f'{controller.service.__blnt__.model_name}.deleted', resp)

        return ControllerRoute(route, '/{value}', web.HttpMethod.DELETE,
                               returns=Returns(self.service.__blnt__.model_name, returns),
                               middlewares=middlewares)



from abc import abstractmethod, ABC
from typing import Callable, Any, Dict, List, Tuple, Optional

from bolinette import blnt, core


class BaseQueryBuilder(ABC):
    def __init__(self, model: 'core.Model', context: 'blnt.BolinetteContext'):
        self.context = context
        self._model = model

    @abstractmethod
    def query(self) -> 'BaseQuery':
        pass

    @abstractmethod
    async def insert_entity(self, values):
        pass

    @abstractmethod
    async def delete_entity(self, entity):
        pass


class BaseQuery(ABC):
    def __init__(self):
        self._filters: Dict[str, Any] = {}
        self._order_by: List[Tuple[Callable[[Any], Any], bool]] = []
        self._offset = 0
        self._limit: Optional[int] = None

    def _base_clone(self, query: 'BaseQuery'):
        query._filters = dict(self._filters)
        query._order_by = dict(self._order_by)
        query._limit = self._limit
        query._offset = self._offset

    @abstractmethod
    def _clone(self) -> 'BaseQuery':
        pass

    def _filter_by_func(self, **kwargs) -> 'BaseQuery':
        for key in kwargs:
            self._filters[key] = kwargs[key]
        return self

    def filter_by(self, **kwargs) -> 'BaseQuery':
        return self._clone()._filter_by_func(**kwargs)

    def _order_by_func(self, function: Callable[[Any], Any], *, desc: bool = False) -> 'BaseQuery':
        self._order_by.append((function, desc))
        return self

    def order_by(self, function: Callable[[Any], Any], *, desc: bool = False) -> 'BaseQuery':
        return self._clone()._order_by_func(function, desc=desc)

    def _offset_func(self, offset: int) -> 'BaseQuery':
        self._offset = offset
        return self

    def offset(self, offset: int) -> 'BaseQuery':
        return self._clone()._offset_func(offset)

    def _limit_func(self, limit: int) -> 'BaseQuery':
        self._limit = limit
        return self

    def limit(self, limit: int) -> 'BaseQuery':
        return self._clone()._limit_func(limit)

    @abstractmethod
    async def all(self):
        pass

    @abstractmethod
    async def first(self):
        pass

    @abstractmethod
    async def get_by_id(self, identifier):
        pass

    @abstractmethod
    async def count(self):
        pass



import sqlalchemy


class DataType:
    def __init__(self, name, sqlalchemy_type):
        self.name = name
        self.sqlalchemy_type = sqlalchemy_type

    def of_type(self, cls):
        return type(self) is cls

    def __repr__(self):
        return self.name


String = DataType('String', sqlalchemy.String)
Integer = DataType('Integer', sqlalchemy.Integer)
Float = DataType('Float', sqlalchemy.Float)
Boolean = DataType('Boolean', sqlalchemy.Boolean)
Date = DataType('Date', sqlalchemy.DateTime)
Email = DataType('Email', sqlalchemy.String)
Password = DataType('Password', sqlalchemy.String)



from datetime import datetime, timedelta

import jwt as py_jwt

from bolinette import blnt
from bolinette.exceptions import UnauthorizedError


class JWT:
    def __init__(self, context: 'blnt.BolinetteContext'):
        self.context = context
        self._access_token_expires = timedelta(days=1)
        self._refresh_token_expires = timedelta(days=30)

    def access_token_expires(self, date):
        return date + self._access_token_expires

    def refresh_token_expires(self, date):
        return date + self._refresh_token_expires

    @property
    def secret_key(self):
        return self.context.env['secret_key']

    def encode(self, payload):
        return py_jwt.encode(payload, self.secret_key, algorithm='HS256').decode('utf-8')

    def decode(self, token):
        return py_jwt.decode(token, self.secret_key, algorithm='HS256')

    def create_access_token(self, date, identity, *, fresh=False):
        payload = {
            'identity': identity,
            'fresh': fresh,
            'expires': str(self.access_token_expires(date))
        }
        return self.encode(payload)

    def create_refresh_token(self, date, identity):
        payload = {
            'identity': identity,
            'expires': str(self.refresh_token_expires(date))
        }
        return self.encode(payload)

    def verify(self, request, *, optional=False, fresh=False):
        now = datetime.utcnow()
        if 'access_token' in request.cookies:
            try:
                payload = self.decode(request.cookies['access_token'])
            except py_jwt.PyJWTError:
                if optional:
                    return None
                raise UnauthorizedError('user.token.invalid')
            expires = datetime.strptime(payload['expires'], '%Y-%m-%d %H:%M:%S.%f')
            if now > expires:
                raise UnauthorizedError('user.token.expired')
            if fresh and not payload['fresh']:
                raise UnauthorizedError('user.token.fresh_required')
            return payload['identity']
        if not optional:
            raise UnauthorizedError('user.unauthorized')
        return None



from bolinette.blnt.database.queries.base import BaseQueryBuilder, BaseQuery
from bolinette.blnt.database.queries.relational import RelationalQueryBuilder, RelationalQuery
from bolinette.blnt.database.queries.collection import CollectionQueryBuilder, CollectionQuery



from bolinette.console import console
from bolinette.bcrypt import bcrypt
import bolinette.defaults
import bolinette.init
from bolinette.commands import command
from bolinette.bolinette import Bolinette



from typing import Any, Dict, Callable, Awaitable

from aiohttp.web_request import Request
from aiohttp.web_response import Response

from bolinette import blnt


class Middleware:
    __blnt__: 'MiddlewareMetadata' = None

    def __init__(self, context: 'blnt.BolinetteContext'):
        self.system_priority = 1
        self.context = context
        self.options = {}

    async def handle(self, request: Request, params: Dict[str, Any],
                     next_func: Callable[[Request, Dict[str, Any]], Awaitable[Response]]):
        return await next_func(request, params)

    def __repr__(self):
        return f'<Middleware {self.__blnt__.name}>'


class MiddlewareMetadata:
    def __init__(self, name: str, priority: int, auto_load: bool, loadable: bool):
        self.name = name
        self.priority = priority
        self.auto_load = auto_load
        self.loadable = loadable


class InternalMiddleware(Middleware):
    def __init__(self, context: 'blnt.BolinetteContext'):
        super().__init__(context)
        self.system_priority = 0



import traceback

from sqlalchemy.exc import SQLAlchemyError

from bolinette import blnt
from bolinette.exceptions import APIError, APIErrors, InternalError


class Transaction:
    def __init__(self, context: 'blnt.BolinetteContext'):
        self.context = context

    async def __aenter__(self):
        await self.context.db.open_transaction()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            await self.context.db.rollback_transaction()
            if not issubclass(exc_type, (APIError, APIErrors)):
                self.context.logger.error(str(exc_val))
                traceback.print_tb(exc_tb)
                if self.context.env['debug']:
                    raise InternalError([str(exc_val)] + traceback.format_list(traceback.extract_tb(exc_tb)))
                raise InternalError('internal.error')
        else:
            try:
                await self.context.db.close_transaction()
            except SQLAlchemyError as err:
                await self.context.db.rollback_transaction()
                raise InternalError([f'global.internal_error:{err}'])



from typing import Dict

from bolinette import core, blnt, mapping
from bolinette.exceptions import InternalError
from bolinette.utils.functions import getattr_, hasattr_


class Mapper:
    def __init__(self, context: 'blnt.BolinetteContext'):
        self.context = context
        self._payloads: Dict[str, mapping.Definition] = {}
        self._responses: Dict[str, mapping.Definition] = {}

    def _get_def(self, collection, model_name, key) -> 'mapping.Definition':
        m = collection.get(model_name)
        if m is None:
            raise InternalError(f'mapping.unknown_model:{model_name}')
        d = m.get(key)
        if d is None:
            raise InternalError(f'mapping.unknown_definition:{key}')
        return d

    def payload(self, model_name: str, key: str):
        return self._get_def(self._payloads, model_name, key)

    def response(self, model_name: str, key: str):
        return self._get_def(self._responses, model_name, key)

    def register(self, model_name: str, model: 'core.Model'):
        def create_defs(collection, params):
            if params is None:
                return
            for param in params:
                if isinstance(param, list):
                    model_key = 'default'
                    payload = param
                else:
                    model_key, payload = param
                definition = mapping.Definition(model_name, model_key)
                for field in payload:
                    definition.fields.append(field)
                if definition.model_name not in collection:
                    collection[definition.model_name] = {}
                collection[definition.model_name][definition.model_key] = definition

        create_defs(self._payloads, model.payloads())
        create_defs(self._responses, model.responses())

    def marshall(self, definition, entity, *, skip_none=False, as_list=False, use_foreign_key=False):
        if not entity:
            return None
        if as_list:
            return [self.marshall(definition, e, skip_none=skip_none, as_list=False, use_foreign_key=use_foreign_key)
                    for e in entity]

        values = {}
        for field in definition.fields:
            if isinstance(field, mapping.Field):
                if field.function is not None:
                    value = field.function(entity)
                else:
                    value = getattr_(entity, field.key, None)
                if field.formatting is not None:
                    value = field.formatting(value)
                if not skip_none or value is not None:
                    values[field.name] = value
            elif isinstance(field, mapping.Reference) and use_foreign_key:
                values[field.foreign_key] = getattr_(entity, field.foreign_key, None)
            elif isinstance(field, mapping.Definition):
                d = self.response(field.model_name, field.model_key)
                attr = None
                if field.function and callable(field.function):
                    attr = field.function(entity)
                elif hasattr_(entity, field.name):
                    attr = getattr_(entity, field.name, None)
                values[field.name] = self.marshall(d, attr, skip_none=skip_none, as_list=False,
                                                   use_foreign_key=use_foreign_key)
            elif isinstance(field, mapping.List):
                d = self.response(field.element.model_name, field.element.model_key)
                values[field.name] = self.marshall(d, getattr_(entity, field.name, None), skip_none=skip_none,
                                                   as_list=True, use_foreign_key=use_foreign_key)
        return values



from bolinette.utils import logger

from bolinette import blnt
from bolinette.mail.providers import Mailgun

_providers = {
    'mailgun': Mailgun
}


class Sender:
    def __init__(self, context: 'blnt.BolinetteContext'):
        self.context = context
        self.provider = None

    def init_app(self):
        provider = self.context.env['mail_provider']
        if provider:
            provider = provider.lower()
            if provider not in _providers:
                logger.warning(f'Unknown "{provider}" mail provider. '
                               f'Available: {", ".join(_providers.keys())}')
            else:
                self.provider = _providers[provider](self.context)

    async def send(self, to, subject, content):
        if self.provider:
            try:
                self.provider.send(to, subject, content)
            except Exception as ex:
                logger.error(str(ex))



from bolinette import types, core, mapping
from bolinette.decorators import mixin


@mixin('historized')
class Historized(core.Mixin):
    @staticmethod
    def columns():
        return {
            'created_on': types.defs.Column(types.db.Date),
            'updated_on': types.defs.Column(types.db.Date),
            'created_by_id': types.defs.Column(types.db.Integer, reference=types.defs.Reference('user', 'id')),
            'updated_by_id': types.defs.Column(types.db.Integer, reference=types.defs.Reference('user', 'id'))
        }

    @staticmethod
    def relationships(model_cls):
        return {
            'created_by': types.defs.Relationship('user', foreign_key=model_cls.created_by_id, lazy=False),
            'updated_by': types.defs.Relationship('user', foreign_key=model_cls.updated_by_id, lazy=False)
        }

    @staticmethod
    def response(model_cls):
        return [
            mapping.Column(model_cls.created_on),
            mapping.Column(model_cls.updated_on),
            mapping.Reference(model_cls.created_by),
            mapping.Reference(model_cls.updated_by),
        ]



import sqlalchemy
from sqlalchemy import orm as sqlalchemy_orm

from bolinette import blnt, core, web
from bolinette.decorators import init_func
from bolinette.exceptions import InitError


@init_func
def init_relational_models(context: blnt.BolinetteContext):
    models = {}
    for model_name, model_cls in blnt.cache.models.items():
        db_key = model_cls.__blnt__.database
        if db_key in context.db:
            if not context.db[db_key].relational:
                continue
            database = context.db[db_key]
        else:
            raise InitError(f'Undefined "{db_key}" database for model "{model_name}"')
        models[model_name] = model_cls(database)
    orm_tables = {}
    orm_cols = {}
    for model_name, model in models.items():
        orm_cols[model_name] = {}
        for att_name, attribute in model.__props__.get_columns().items():
            attribute.name = att_name
            ref = None
            if attribute.reference:
                ref = sqlalchemy.ForeignKey(f'{attribute.reference.model_name}.{attribute.reference.column_name}')
            orm_cols[model_name][att_name] = sqlalchemy.Column(
                att_name, attribute.type.sqlalchemy_type, ref, default=attribute.default,
                primary_key=attribute.primary_key, nullable=attribute.nullable, unique=attribute.unique)
        orm_tables[model_name] = sqlalchemy.Table(model_name,
                                                  model.__props__.database.base.metadata,
                                                  *(orm_cols[model_name].values()))

    for model_name, model in models.items():
        orm_defs = {}
        for att_name, attribute in model.__props__.get_relationships().items():
            kwargs = {}
            attribute.name = att_name
            if attribute.backref:
                kwargs['backref'] = sqlalchemy_orm.backref(attribute.backref.key, lazy=attribute.backref.lazy)
            if attribute.foreign_key:
                kwargs['foreign_keys'] = orm_cols[model_name][attribute.foreign_key.name]
            if attribute.remote_side:
                kwargs['remote_side'] = orm_cols[model_name][attribute.remote_side.name]
            if attribute.secondary:
                kwargs['secondary'] = orm_tables[attribute.secondary]
            orm_defs[att_name] = sqlalchemy_orm.relationship(attribute.model_name,  lazy=attribute.lazy, **kwargs)

        orm_defs['__table__'] = orm_tables[model_name]
        orm_model = type(model_name, (model.__props__.database.base,), orm_defs)

        for att_name, attribute in model.__props__.get_properties().items():
            setattr(orm_model, att_name, property(attribute.function))

        context.add_model(model_name, model)
        context.add_table(model_name, orm_model)


@init_func
def init_collection_models(context: blnt.BolinetteContext):
    models = {}
    for model_name, model_cls in blnt.cache.models.items():
        db_key = model_cls.__blnt__.database
        if db_key in context.db:
            if context.db[db_key].relational:
                continue
            database = context.db[db_key]
        else:
            raise InitError(f'Undefined "{db_key}" database for model "{model_name}"')
        models[model_name] = model_cls(database)
    for model_name, model in models.items():
        for att_name, attribute in model.__props__.get_columns().items():
            attribute.name = att_name
        context.add_model(model_name, model)


@init_func
def init_repositories(context: blnt.BolinetteContext):
    for model_name, model in context.models:
        context.add_repo(model_name, core.Repository(model_name, model, context))


@init_func
def init_mappings(context: blnt.BolinetteContext):
    for model_name, model in context.models:
        context.mapper.register(model_name, model)


@init_func
def init_services(context: blnt.BolinetteContext):
    for service_name, service_cls in blnt.cache.services.items():
        context.add_service(service_name, service_cls(context))


@init_func
def init_controllers(context: blnt.BolinetteContext):
    def _init_sys_middleware(_route: 'web.ControllerRoute'):
        sys_mdw = [m.__blnt__.name for m in blnt.cache.middlewares.values() if m.__blnt__.auto_load]
        if _route.expects is not None:
            model = _route.expects.model
            key = _route.expects.key if _route.expects.key is not None else 'default'
            cmd = f'blnt_payload|model={model}|key={key}'
            if _route.expects.patch:
                cmd += '|patch'
            sys_mdw.append(cmd)
        else:
            sys_mdw.append('blnt_payload')
        if _route.returns is not None:
            model = _route.returns.model
            key = _route.returns.key if _route.returns.key is not None else 'default'
            cmd = f'blnt_response|model={model}|key={key}'
            if _route.returns.as_list:
                cmd += '|as_list'
            if _route.returns.skip_none:
                cmd += '|skip_none'
            sys_mdw.append(cmd)
        else:
            sys_mdw.append('blnt_response')
        return sys_mdw

    def _add_route(_controller: web.Controller, _route: web.ControllerRoute):
        path = f'{_controller.__blnt__.namespace}{_controller.__blnt__.path}{_route.path}'
        _route.controller = _controller
        _route.init_middlewares(context, _controller.__blnt__.middlewares, _init_sys_middleware(_route))
        context.resources.add_route(path, _controller, _route)
        if _route.inner_route is not None:
            _add_route(_controller, _route.inner_route)

    for controller_name, controller_cls in blnt.cache.controllers.items():
        controller = controller_cls(context)
        for _, route in controller.__props__.get_routes().items():
            _add_route(controller, route)
        if isinstance(controller, web.Controller):
            for route in controller.default_routes():
                _add_route(controller, route)
        context.add_controller(controller_name, controller)


@init_func
def init_topics(context: blnt.BolinetteContext):
    context.sockets.init_socket_handler()
    for topic_name, topic_cls in blnt.cache.topics.items():
        topic = topic_cls(context)
        context.sockets.add_topic(topic_name, topic)
        for channel_name, channel in topic.__props__.get_channels().items():
            context.sockets.add_channel(topic_name, channel)



import random
import string
from datetime import datetime

from bolinette import blnt, web
from bolinette.decorators import controller, get, post, patch, delete
from bolinette.defaults.services import UserService, RoleService
from bolinette.exceptions import UnprocessableEntityError, EntityNotFoundError


@controller('user', '/user')
class UserController(web.Controller):
    @property
    def user_service(self) -> UserService:
        return self.context.service('user')

    @property
    def role_service(self) -> RoleService:
        return self.context.service('role')

    def default_routes(self):
        return [
            self.defaults.get_all('private', middlewares=['auth|roles=admin']),
            self.defaults.get_one('private', key='username', middlewares=['auth|roles=admin'])
        ]

    def _create_tokens(self, resp, user, *, set_access, set_refresh, fresh):
        now = datetime.utcnow()
        if set_access:
            access_token = self.context.jwt.create_access_token(now, user.username, fresh=fresh)
            resp.cookies.append(web.Cookie('access_token', access_token,
                                           expires=self.context.jwt.access_token_expires(now), path='/'))
        if set_refresh:
            refresh_token = self.context.jwt.create_refresh_token(now, user.username)
            resp.cookies.append(web.Cookie('refresh_token', refresh_token,
                                           expires=self.context.jwt.refresh_token_expires(now),
                                           path='/api/user/refresh'))

    @get('/me', returns=web.Returns('user', 'private'), middlewares=['auth|fresh'])
    async def me(self, current_user):
        return self.response.ok('OK', current_user)

    @get('/info', returns=web.Returns('user', 'private'), middlewares=['auth'])
    async def info(self, current_user):
        return self.response.ok('OK', current_user)

    @post('/login', expects=web.Expects('user', 'login'), returns=web.Returns('user', 'private'))
    async def login(self, payload):
        username = payload['username']
        password = payload['password']
        try:
            user = await self.user_service.get_by_username(username)
        except EntityNotFoundError:
            return self.response.unauthorized('user.login.wrong_credentials')
        if user is not None:
            if self.user_service.check_password(user, password):
                resp = self.response.ok('user.login.success', user)
                self._create_tokens(resp, user, set_access=True, set_refresh=True, fresh=True)
                return resp
        return self.response.unauthorized('user.login.wrong_credentials')

    @post('/logout')
    async def logout(self):
        resp = self.response.ok('user.logout.success')
        resp.cookies.append(web.Cookie('access_token', None, delete=True, path='/'))
        resp.cookies.append(web.Cookie('refresh_token', None, delete=True, path='/api/user/refresh'))
        return resp

    @post('/token/refresh', middlewares=['auth'])
    async def refresh(self, current_user):
        resp = self.response.ok('user.token.refreshed')
        self._create_tokens(resp, current_user, set_access=True, set_refresh=False, fresh=False)
        return resp

    @post('/register', expects=web.Expects('user', 'register'), returns=web.Returns('user', 'private'))
    async def register(self, payload):
        if blnt.init.get('ADMIN_REGISTER_ONLY', True):
            raise UnprocessableEntityError('global.register.admin_only')
        user = await self.user_service.create(payload)
        resp = self.response.created('user.registered', user)
        self._create_tokens(resp, user, set_access=True, set_refresh=True, fresh=True)
        return resp

    @post('/register/admin', expects=web.Expects('user', 'admin_register'),
          returns=web.Returns('user', 'private'), middlewares=['auth|roles=admin'])
    async def admin_register(self, payload):
        # send_mail = payload.pop('send_mail')
        payload['password'] = ''.join(random.choices(string.ascii_lowercase, k=32))
        user = await self.user_service.create(payload)
        # if send_mail:
        #     await mail.sender.send(payload['email'], 'Welcome!', 'Welcome to Bolinette!')
        return self.response.created('user.registered', user)

    @patch('/me', expects=web.Expects('user', 'register', patch=True), returns=web.Returns('user', 'private'),
           middlewares=['auth|fresh'])
    async def update_user(self, payload, current_user):
        user = await self.user_service.patch(current_user, payload)
        resp = self.response.ok('user.updated', user)
        self._create_tokens(resp, user, set_access=True, set_refresh=True, fresh=True)
        return resp

    @post('/{username}/roles', expects=web.Expects('role'), returns=web.Returns('user', 'private'),
          middlewares=['auth|roles=admin'])
    async def add_user_role(self, match, payload):
        user = await self.user_service.get_by_username(match['username'])
        role = await self.role_service.get_by_name(payload['name'])
        await self.user_service.add_role(user, role)
        return self.response.created(f'user.roles.added:{user.username}:{role.name}', user)

    @delete('/{username}/roles/{role}', returns=web.Returns('user', 'private'), middlewares=['auth|roles=admin'])
    async def delete_user_role(self, match, current_user):
        user = await self.user_service.get_by_username(match['username'])
        role = await self.role_service.get_by_name(match['role'])
        await self.user_service.remove_role(current_user, user, role)
        return self.response.ok(f'user.roles.removed:{user.username}:{role.name}', user)

    @post('/picture', returns=web.Returns('user', 'private'), middlewares=['auth'])
    async def upload_profile_picture(self, current_user, payload):
        picture = payload['file']
        user = await self.user_service.save_profile_picture(current_user, picture)
        return self.response.ok(f'user.picture.uploaded', user)



import getpass

from bolinette.utils import console

import bolinette
from bolinette.commands import command
from bolinette.exceptions import ParamConflictError, EntityNotFoundError


@command('create_user')
def create_user(blnt: 'bolinette.Bolinette', *, username, email, roles):
    user_service = blnt.context.service('user')
    role_service = blnt.context.service('role')
    while True:
        password = getpass.getpass('Choose password: ')
        password2 = getpass.getpass('Confirm password: ')
        if password == password2:
            break
        console.error('Passwords don\'t match')
    with blnt.app.app_context():
        user_roles = []
        if user_roles is not None:
            for role_name in [r.strip() for r in roles.split(',')]:
                try:
                    user_roles.append(role_service.get_by_name(role_name))
                except EntityNotFoundError:
                    console.error(f'Role "{role_name}" does not exist')
                    exit(1)
        try:
            user = user_service.create({
                'username': username,
                'password': password,
                'email': email
            })
            for role in user_roles:
                user.roles.append(role)
        except ParamConflictError as ex:
            console.error(f'Conflict: {ex.message.split(":")[1]} already exists')
            exit(1)
        blnt.context.db.session.commit()



from typing import List

from bolinette import blnt, core
from bolinette.blnt.objects import Pagination, PaginationParams, OrderByParams
from bolinette.blnt.database.queries import RelationalQueryBuilder, BaseQuery, CollectionQueryBuilder
from bolinette.exceptions import ParamConflictError, APIErrors
from bolinette.utils.functions import setattr_, getattr_


class Repository:
    def __init__(self, name: str, model: 'core.Model', context: 'blnt.BolinetteContext'):
        self.name = name
        self.model = model
        self.context = context
        self._query_builder = self._get_query_builder(model, context)

    @staticmethod
    def _get_query_builder(model: 'core.Model', context: 'blnt.BolinetteContext'):
        db = context.db[model.__blnt__.database]
        if db.relational:
            return RelationalQueryBuilder(model, context)
        return CollectionQueryBuilder(model, context)

    def __repr__(self):
        return f'<Repository {self.name}>'

    def query(self):
        return self._query_builder.query()

    async def get_all(self, pagination: PaginationParams = None, order_by: List[OrderByParams] = None):
        if order_by is None:
            order_by = []
        query = self.query()
        if len(order_by) > 0:
            query = await self._build_order_by(query, order_by)
        if pagination is not None:
            return await self._paginate(query, pagination)
        return await query.all()

    async def get(self, identifier):
        return await self.query().get_by_id(identifier)

    async def get_by(self, key, value):
        return await self.query().filter_by(**{key: value}).all()

    async def get_first_by(self, key, value):
        return await self.query().filter_by(**{key: value}).first()

    async def create(self, values):
        filtered = await self._validate_model(values)
        entity = await self._query_builder.insert_entity(filtered)
        return entity

    async def update(self, entity, values):
        await self._map_model(entity, values)
        return entity

    async def patch(self, entity, values):
        await self._map_model(entity, values, patch=True)
        return entity

    async def delete(self, entity):
        return await self._query_builder.delete_entity(entity)

    async def _build_order_by(self, query: BaseQuery, params):
        order_by_query = []
        for col_name, way in params:
            if hasattr(self._table, col_name):
                column = getattr(self._table, col_name)
                if way:
                    order_by_query.append(column)
                else:
                    order_by_query.append(core.functions.desc(column))
        return query.order_by(*order_by_query)

    async def _paginate(self, query, pagination: PaginationParams):
        page = pagination.page
        per_page = pagination.per_page
        total = await query.count()
        items = await query.offset(page * per_page).limit(per_page).all()
        return Pagination(items, page, per_page, total)

    async def _validate_model(self, values: dict):
        api_errors = APIErrors()
        for column in self.model.__props__.get_columns().values():
            key = column.name
            if column.primary_key:
                continue
            if key in values:
                value = values.get(key)
                if column.unique:
                    if await self.get_first_by(column.name, value) is not None:
                        api_errors.append(ParamConflictError(key, value))
        if api_errors:
            raise api_errors
        return values

    async def _map_model(self, entity, values, patch=False):
        api_errors = APIErrors()
        for _, column in self.model.__props__.get_columns().items():
            key = column.name
            if column.primary_key or (key not in values and patch):
                continue
            original = getattr_(entity, key, None)
            new = values.get(key, None)
            if original == new:
                continue
            if column.unique and new is not None:
                if await self.get_first_by(column.name, new) is not None:
                    api_errors.append(ParamConflictError(key, new))
            setattr_(entity, key, new)
        if api_errors:
            raise api_errors



import bcrypt as py_bcrypt


class Bcrypt:
    def hash_password(self, password):
        return py_bcrypt.hashpw(password.encode(), py_bcrypt.gensalt())

    def check_password(self, hashed, password):
        return py_bcrypt.checkpw(password.encode(), hashed)


bcrypt = Bcrypt()
